/**
 * Copyright (C) 2022 Chris Noxz
 * Author(s): Chris Noxz <chris@noxz.tech>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef ACST_H
#define ACST_H

#include <stdbool.h>        // for bool
#include <stdio.h>          // for printf
#include <stdlib.h>         // for exit, EXIT_SUCCESS, size_t

#define BUFSZ               8192    /* buffer size to use in hash computation */
#define SHA256_BYTES        32      /* byte size of SHA256 hash */
#define CSSZ                64      /* size of SHA256 hash stored as hex */
#define TSSZ                32      /* allocation size for timespan holder */
#define XATTR_CS            "user.acst.cs"  /* name of checksum xattr */
#define XATTR_TS            "user.acst.ts"  /* name of timestamp xattr */
#define FRMT_TS             "%010llu.%09lu" /* format of timestamp xattr */

/* string values used in program, could be changes for language compliance */
#define STR_CMP_FRMT        " stored: %s "FRMT_TS"\n"                       \
                            " actual: %s "FRMT_TS"\n"
#define STR_ERR_OOM         "Error: out of memory\n"
#define STR_ERR_ABNO        "Error: abnormal changes detected \"%s\"\n"
#define STR_ERR_OPNF        "Error: could not open file \"%s\"\n"
#define STR_ERR_OPND        "Error: could not open directory \"%s\"\n"
#define STR_ERR_DDEV        "Error: \"%s\" belong to a different devices than \"%s\"\n"
#define STR_ERR_REGF        "Error: not a regular file \"%s\"\n"
#define STR_ERR_HASH        "Error: could not compute hash \"%s\"\n"
#define STR_ERR_XARM        "Error: could not remove extended attributes from file \"%s\": %s\n"
#define STR_ERR_XAWR        "Error: could not write extended attributes to file \"\%s\"\n"
#define STR_ERR_RECU        "Error: path is a directory, did you mean to use the -r option? \"%s\"\n"
#define STR_OUT_STAT        "<%s> %s\n"
#define STR_OUT_NEW         "new"
#define STR_OUT_OK          "ok"
#define STR_OUT_TOUC        "hash ok"
#define STR_OUT_OUTD        "outdated"
#define STR_OUT_BACK        "backdated"
#define STR_OUT_CORR        "corrupt"
#define STR_OUT_MALF        "malformed"
#define STR_OUT_DISR        "disrupted"
#define STR_OUT_REMO        "xattr removed"

#define USAGE(code)                                                         \
        {                                                                   \
          printf("Usage: %s [OPTION]... <FILE>...\n", arg.prg);             \
          if ((code) == EXIT_SUCCESS)                                       \
            printf(                                                         \
    "\n"                                                                    \
    "Tracks changes and corruption in files using xattr-based checksums.\n" \
    "\n"                                                                    \
    "Positional arguments:\n"                                               \
    "  FILE                  file(s) to track\n"                            \
    "\n"                                                                    \
    "Optional arguments:\n"                                                 \
    "  -h                    show this help message and exit\n"             \
    "  -m                    summarize information at end of execution\n"   \
    "  -n                    don't create or update any file attributes\n"  \
    "  -q                    quiet mode, lowers verbosity\n"                \
    "  -r                    process files and directories recursively\n"   \
    "  -V                    output version information and exit\n"         \
    "  -x                    remove extended attributes from file(s)\n"     \
    "\n"                                                                    \
    "Full documentation <https://noxz.tech/software/acst>\n"                \
            );                                                              \
          exit(code);                                                       \
        }
#define VER()                                                               \
        {                                                                   \
          printf("%s " VERSION "\n", arg.prg);                              \
          exit(EXIT_SUCCESS);                                               \
        }
#define SUMMERIZE()                                                         \
        {                                                                   \
          if (arg.remove)                                                   \
            printf(         "\n"                                            \
                            "Total files processed      : %d\n"             \
                            "  Xattr removed            : %d\n"             \
                            "Total errors               : %d\n"             \
                            "  Opening files            : %d\n"             \
                            "  Non-regular files        : %d\n"             \
                            "  Xattr operations         : %d\n"             \
                            "  Generic                  : %d\n"             \
              , cnt.total, cnt.ok                                           \
              , cnt.errs, cnt.errOpening, cnt.errNotRegular                 \
              , cnt.errWritingXattr, cnt.errGeneric);                       \
          else                                                              \
            printf(         "\n"                                            \
                            "Total files processed      : %d\n"             \
                            "  New                      : %d\n"             \
                            "  Ok                       : %d\n"             \
                            "  Hash ok                  : %d\n"             \
                            "  Outdated                 : %d\n"             \
                            "  Backdated                : %d\n"             \
                            "  Corrupt                  : %d\n"             \
                            "  Malformed                : %d\n"             \
                            "  Disrupted                : %d\n"             \
                            "Total errors               : %d\n"             \
                            "  Opening files            : %d\n"             \
                            "  Non-regular files        : %d\n"             \
                            "  Xattr operations         : %d\n"             \
                            "  Generic                  : %d\n"             \
              , cnt.total, cnt.new, cnt.ok, cnt.touched                     \
              , cnt.outdated, cnt.backdated, cnt.corrupt                    \
              , cnt.malformed, cnt.disrupted                                \
              , cnt.errs, cnt.errOpening, cnt.errNotRegular                 \
              , cnt.errWritingXattr, cnt.errGeneric);                       \
        }

/* helper variables used by various functions */
static char ts[TSSZ];                               /* timestamp holder */
static const char hextab[]  = "0123456789abcdef";   /* hex digit table */
static const char zSHA256[] = "0000000000000000"    /* zeroed out SHA256 */
                              "0000000000000000"    /* why 4 lines?... */
                              "0000000000000000"
                              "0000000000000000";   /* ...it looks better! */

enum Error {
	ER_NOT_REGULAR,         /* not a regular file error */
	ER_OPENING,             /* error opening file or directory */
	ER_XATTR_OPERATION,     /* error when performing xattr operations */
	ER_GENERIC,             /* generic error */
	ER_FATAL                /* fatal error, should result in program ending */
};

enum FileState {
	FS_OK,                  /* checksum and mtime both match */
	FS_DISRUPTED,           /* file was changed during hash computation */
	FS_CORRUPT,             /* checksum differs while mtime matches */
	FS_MALFORMED,           /* xattrs cannot be read in its intended format */
	FS_TOUCHED,             /* checksum matches but mtime doesn't */
	FS_NEW,                 /* file has no prior checksum or mtime stored */
	FS_OUTDATED,            /* checksum differs and mtime is newer */
	FS_BACKDATED,           /* checksum differs and mtime is older */
	FS_REMOVED,             /* xattrs have been removed from file */
	FS_REMOVED_ERROR,       /* removal of xattrs failed */
	FS_ERROR                /* general error */
};

/* extended attribute/metadata structure */
typedef struct {
	unsigned long long s;   /* seconds */
	unsigned long ns;       /* nanoseconds */
	char cs[CSSZ + 1];      /* checksum */
	int tcmp;               /* timespan comparison holder */
} xa_t;

/* arguments collected from command line */
struct Arguments {
	const char *prg;        /* program name */
	bool dryrun;            /* make a dry run */
	int quiet;              /* level of quietness */
	bool recursive;         /* indicate to open and walk directories */
	bool remove;            /* remove xattrs */
	bool summarize;         /* show summery at end of program */
} arg;

/* counters */
struct Counters {
	int errs;               /* all errors */
	int errNotRegular;      /* not a regular file errors  */
	int errOpening;         /* errors opening file or directory */
	int errWritingXattr;    /* errors when performing xattr operations */
	int errGeneric;         /* generic errors */

	int total;              /* all non error counters */
	int ok;                 /* checksum and mtime both match */
	int disrupted;          /* file was changed during hash computation */
	int corrupt;            /* checksum differs while mtime matches */
	int malformed;          /* xattrs cannot be read in its intended format */
	int outdated;           /* checksum differs and mtime is newer */
	int backdated;          /* checksum differs and mtime is older */
	int touched;            /* checksum matches but mtime doesn't */
	int new;                /* file has no prior checksum or mtime stored */
} cnt;

/**
 * Converts binary data into hex format.
 *
 * @param bin               The binary data
 * @param len               The size of the data
 * @param hex               The result of the converted data
 */
static void bin2hex(unsigned char *bin, size_t len, char *hex);

/**
 * Retrieves and reports error types and messages. If error type is of fatal
 * nature, the program will exit with EXIT_FAILURE.
 *
 * @param er                The type of error (see enum Error above)
 * @param fmt               The output format of the error message
 * @param ...               Various arguments to output in the message
 */
static void error(enum Error er, const char *fmt, ...);

/**
 * Final stage of processing a file which returns the state of the file being
 * processed meanwhile metadata holders are being populated.
 *
 * @param fd                File descriptor of the file
 * @param xa_s              Metadata holder of (stored) extended attributes
 * @param xa_a              Metadata holder of (actual) extended attributes
 *
 * @returns                 The file state of the file being processed
 */
static enum FileState file_state(int fd, xa_t *xa_s, xa_t *xa_a);

/**
 * Processes arguments (file names) either from command line or from directory
 * output.
 *
 * @param fn                File name of the file
 */
static void process_argument(const char *fn);

/**
 * Processes directories if stated to do so (see arg.recursive) by looping
 * through their content (children).
 *
 * @param fn                File name of the directory
 */
static void process_directory(const char *fn);

/**
 * Processes files and returns either error messages or other output if stated
 * to do so (see arg.quiet) based on their file state (see enum FileState
 * above).
 *
 * @param fn                File name of the file
 */
static void process_file(const char *fn);

/**
 * Reports on files and their file state.
 *
 * @param fn                File name of the file
 * @param st                The state of the file
 */
static void report(const char *fn, enum FileState st);

/**
 * Compute and retrieve metadata for the checksum based on file content
 * together with the date modified.
 *
 * @param fd                File descriptor of the file
 * @param xa                Metadata holder of extended attributes
 *
 * @returns                 True if no error occurred
 * @returns                 False if an error occurred
 */
static bool xa_compute(int fd, xa_t *xa);

/**
 * Computes the hash checksum of a file based on the SHA256 hashing algorithm.
 *
 * @param fd                File descriptor of the file
 * @param cs                Pointer for the checksum
 */
static void xa_hash(int fd, char *cs);

/**
 * Reads and retrieves extended attributes as metadata from a file.
 *
 * @param fd                File descriptor of the file
 * @param xa                Metadata holder of extended attributes
 *
 * @returns                 True if no error occurred
 * @returns                 False if an error occurred (malformed data)
 */
static bool xa_read(int fd, xa_t *xa);

/**
 *  Writes, or removes if stated (see arg.remove), extended attributes as
 *  metadata to a file.
 *
 * @param fd               File descriptor of the file
 * @param xa                Metadata holder of extended attributes
 *
 * @returns                 True if no error occurred
 * @returns                 False if an error occurred (write error)
 */
static bool xa_write(int fd, const xa_t *xa);

#endif /* !ACST_H */
